import numpy as np
import scipy.integrate
import matplotlib.pyplot as plt


class Parameter:
    C_IN_G= 5 # g/L
    C_IN_A= 1.5 # g/L
    def __init__(self):
        self.Kg=0.01 # g/L
        self.Ka=self.Kg # Ka and Kg have little influence on overall shape with
                        # H1=.1, H2=10, mu_max_a=.05, mu_max_g=.5
        self.D=0.5
        self.H1=10
        self.H2=0.1
        self.mu_max_g= 0.5 # h^-1
        self.mu_max_a= 5
        # self.mu_max_a= 0.1*self.mu_max_g
        self.Yxg=0.5 # g/g
        self.Yxa=self.Yxg # Yxa and Yxg have little influence on overall
                          # shape with H1=.1, H2=10, mu_max_a=.05, mu_max_g=.5

def update(t, x, p):
    """ update of the ode according to Kussel with Monod growth in chemostat
    x0 = [arabinose cells]
    x1 = [glucose cells]
    x2 = [arabinose]
    x3 = [glucose]
    """
    dx=np.zeros_like(x)

    # should be between .1 and .0001
    # https://bionumbers.hms.harvard.edu/search.aspx?trm=e.+coli+monod+constant
    mu_a=p.mu_max_a * x[2]/(p.Ka+x[2])
    mu_g=p.mu_max_g * x[3]/(p.Kg+x[3])

    dx[0]=(mu_a-p.D-p.H1)*x[0] + p.H2 * x[1]
    dx[1]=(mu_g-p.D-p.H2)*x[1] + p.H1 * x[0]
    dx[2]=(p.C_IN_A-x[2])*p.D - x[0] * mu_a/p.Yxa
    dx[3]=(p.C_IN_G-x[3])*p.D - x[1] * mu_g/p.Yxg

    # for i, x in enumerate(x):
    #     if (x<0) and (dx[i]<0):
    #         dx[i]=0

    return dx

def integrate(p, init, title):
    r=scipy.integrate.solve_ivp(fun=update, t_span=[0,80],
                                    y0=[0,init,0,0], args=(p,))

    fig, ax=plt.subplots(1,2)

    ax[0].plot(r.t, r.y[0,:], label='arabinose cells')
    ax[0].plot(r.t, r.y[1,:], label='glucose cells')
    # ax[0].legend()

    ax[1].plot(r.t, r.y[2,:], label='arabinose')
    ax[1].plot(r.t, r.y[3,:], label='glucose')
    # ax[1].legend()

    fig.suptitle(title)

    # fig.savefig('./figures/'+title+'.pdf', bbox_inches='tight')
    # plt.close(fig)

    print(title)


if __name__ == '__main__':


    jump=np.logspace(-2,2,5, base=10)
    print(jump)
    # jump=np.linspace(-3,3,3)

    init=2e6 * 1e6 * 500 * 1e-15# cells/uL * uL/L * fg/cell * g/fg
    # https://bionumbers.hms.harvard.edu/bionumber.aspx?id=111211&ver=4&trm=coli+dry+mass&org=
    # init=2e6 * 1e6 * 280 * 1e-15# cells/uL * uL/L * fg/cell * g/fg
    # https://bionumbers.hms.harvard.edu/search.aspx?trm=mass+e+coli

    p=Parameter()

    # integrate(p, init, 'base')

    new_p=Parameter()

    for k,H1 in enumerate(jump):
        new_p.H1=H1*p.H1
        for l,H2 in enumerate(jump):
            new_p.H2=H2*p.H2
            title=( 'H1-'+str(k)+
                    '-H2-'+str(l))
            integrate(new_p, init, title)

    plt.show()

    # for k,mu_max_g in enumerate(jump):
    #     new_p.mu_max_g=mu_max_g*p.mu_max_g
    #     for l,mu_max_a in enumerate(jump):
    #         new_p.mu_max_a=mu_max_a*p.mu_max_a
    #         title=( 'mu_max_g-'+str(k)+
    #                 '-mu_max_a-'+str(l))
    #         integrate(new_p, init, title)
    #
    # plt.show()

    for k,Kg in enumerate(jump):
        new_p.Kg=Kg*p.Kg
        for l,Ka in enumerate(jump):
            new_p.Ka=Ka*p.Ka
            title=( 'Kg-'+str(k)+
                    '-Ka-'+str(l))
            integrate(new_p, init, title)

    plt.show()

    # for k,Yxg in enumerate(jump):
    #     new_p.Yxg=Yxg*p.Yxg
    #     for l,Yxa in enumerate(jump):
    #         new_p.Yxa=Yxa*p.Yxa
    #         title=( 'Yxg-'+str(k)+
    #                 '-Yxa-'+str(l))
    #         integrate(new_p, init, title)

    # plt.show()
