import numpy as np
import os.path
import pickle
import datetime
import matplotlib.pyplot as plt

class MetropolisHastings:
    """ Metropolis-Hastings (MH) algorithm class
    Requires a "model" object with:
        * a clone method to clone the object without its results
        * a log_evaluate method evaluating the probability of the object (in
            log space) or something proportional. For efficiency, should save
            the evaluation
        * a vary method to transform the object into a new one, for efficiency
            it should probably compute and save the transition and back-
            transition probabilities
        * get_q_forward and get_q_backward methods to output the aforementioned
            transition probabilities. get_q_forward=q(new model | old model)
        * an aggregate method allows to save the 'changing' part of the model.
            The aggregate is expected to be an array with the estimator first
            and the varying parameters afterwards
        * based on an aggregate and a base model, a "degregate" method should
            allow restoration of the old model

    Aggregate is the results: first column is the estimator, the rest is the
    parameters
    """

    def __init__(self, model, a=None):
        # model is the model to optimize
        # with initial parameters and all
        # should save an array of aggregate models as well, will depend on the
        # read method
        # a is the initial aggregate if available
        self.mod=model
        if a is None:
            a=model.aggregate()
        self.aggregate=np.empty((1, a.shape[-1]))
        self.aggregate[0,:]=a

    def clone(self):
        # keep the same mod and not a copy as variations are made on copies
        return MetropolisHastings(self.mod, self.aggregate.copy())

    # @profile
    def run(self, nb_iter=1, rand=None):
        """Run the metropolis hastings algorithm for nb_iter times
        Can add iterations to the previous ones too
        """
        # model is always the last iteration of the model
        # rand is a random array of size nb_iter, can be added for efficiency
        if rand is None:
            rand=np.log(np.random.rand(nb_iter))

        s=self.aggregate.shape[0]
        new=np.empty((s+nb_iter,self.aggregate.shape[1]))
        new[0:s,:]=self.aggregate
        self.aggregate=new
        for i in range(0, nb_iter):
            print("from "+str(s+i-1)+" to "+str(s+i))
            print(str(datetime.datetime.now()))
            tr=self.mod.clone() # tr for the transient value
            tr.vary()
            # a=transient.evaluate()/self.model.evaluate()
            #   *transient.get_q_backward()/transient.get_q_forward()
            # in log scale to reduce risks of numerical overflow
            a=( tr.log_evaluate()    - self.mod.log_evaluate()+
                tr.get_q_backward() - tr.get_q_forward())


            # print(np.exp(a))
            # print(a)
            # print(rand[i])
            # print(tr.log_evaluate())
            # print(self.mod.log_evaluate())
            # print(tr.get_q_backward())
            # print(tr.get_q_forward())

            if rand[i]<a: # select tr with probability a or always tr if a>1
                self.mod=tr
            self.aggregate[s+i,:]=self.mod.aggregate()
            print(self.aggregate[s+i,:])

            if (s+i)%50==0:
                if s+i>100:
                    # idem is the proportion of times it didn't change in the
                    # last 50 times
                    idem=(np.sum(np.all(self.aggregate[s+i-50:s+i,:]
                                    -self.aggregate[s+i-51:s+i-1,:]==0, axis=1))
                                    /50)
                    if idem>=0.9 and self.mod.beta>10:
                    # never changed -> increase temperature
                    # = decrease beta
                        self.mod.beta/=1.5
                    elif idem<=0.1:
                        self.mod.beta*=1.5

                    print("idem")
                    print(idem)



    def save(self, file_name):
        """Save minimal data to be able to reproduce the results
        Keep the estimator for convenience
        """
        # considering the aggregate does not contain the data that does not
        # change from one model to the next, need to save one model as well
        self.mod=self.mod.clone(full=False) # since clone does not keep the results of
                                  # the model and the results can be reproduced
                                  # based on the content of the aggregate...
        # path=os.path.abspath(os.path.expanduser(file_name))
        # filehandler=open(path, 'w')
        # pickle.dump(self, filehandler)
        # filehandler.close()
        with open(file_name+'.pkl', 'wb') as output:
            pickle.dump(self, output, pickle.HIGHEST_PROTOCOL)
            output.close()

    def read(file_name):
        """Read a saved file and make a model out of it"""
        path=os.path.abspath(os.path.expanduser(file_name))
        with open(path+'.pkl', "rb") as input:
            f=pickle.load(input)
            input.close()
            return f


    def show(self, pre='', title=True):
        """Make nice graphs to show how the algorithm evolved"""
        # first show the evolution of the estimator
        plt.figure(0)
        plt.title(pre+' estimator')
        plt.plot(self.aggregate[:,0])

        # show the scatter plot of estimator vs parameters
        for i in range(1, self.aggregate.shape[1]):
            plt.figure(i)
            name, lin=self.mod.info_num(i)

            if title:
                plt.title(pre+" parameter " + name +' '+str(i-1))

            if lin:
                plt.plot(self.aggregate[:,i], self.aggregate[:,0], 'x')
            else:
                plt.semilogx(self.aggregate[:,i], self.aggregate[:,0], 'x')

    def get_mins(self, nb):
        """get an array of the nb minimum estimator models"""
        cp=np.copy(self.aggregate[:,0])
        models=np.empty(nb, dtype=object)
        for i in range(nb):
            index=np.argmin(cp)
            m=cp[index]
            models[i]=self.mod.clone()
            models[i].degregate(self.aggregate[index,:])
            while index<cp.shape[0] and cp[index]==m:
                cp[index]=np.inf
                index+=1
        return models
