import numpy as np
import os.path
import os
import multiprocessing.pool
from segmodels import Model

class MetroWrapper:
    """ Metropolis-Hastings model wrapper class
    This class is a wrapper for a segmodel and its comparison to reality
    It provides all the required methods for the Metropolis-Hastings algorithm
    """

    def __init__(self, model, segdata, beta=1):
        """ initialize the object with a model and data to compare it with
        segdata is an iterable of segdata objects
        """
        self.BASE_MOD=model
        # base model should never be run !
        self.data=segdata
        self.beta=beta

    def clone(self, full=True):
        """clone only the important part of the model"""
        # BASE_MOD needs to be cloned since it'll be varied
        # data is always the same anyways
        # beta is independent
        m= MetroWrapper(self.BASE_MOD.clone(), self.data, beta=self.beta)
        if hasattr(self, "models"):
            if full:
                m.models=np.empty_like(self.models)
                for i, mod in enumerate(self.models):
                    m.models[i]=mod.clone()
        return m

    # @profile
    def _log_async(self,i):
        """wrapper for log_evaluate to be asynchronous"""
        s= self.models[i].compare(self.data[i])
        return s

    def run_once(self):
        self.set_models()
        for i,m in enumerate(self.models):
            m.compare(self.data[i])

    def set_models(self):
        if not hasattr(self, 'models'):
            self.models=[]
            for data in self.data:
                # put the right regulation in the model
                m=self.BASE_MOD.clone()
                m.set_regulation(data)
                self.models.append(m)
    # @profile
    def log_evaluate(self, offset=0):
        """evaluate the log of the probability to have the object
        the offset may be added to the result to avoid numerical instabilities
        """
        if (hasattr(self, "_log_eval") and
                self._log_eval is not None and
                not np.isnan(self._log_eval)):
            return self._log_eval+offset
        else:
            # comparison is to minimize
            # create models specific to the data
            self.set_models()
            # make the evaluation
            log_eval=0
            weight=0
            with multiprocessing.pool.Pool(
                    min(
                        os.cpu_count()-1,self.data.shape[0])
                    ) as p:
                m=p.map(self._log_async, range(self.data.shape[0]))
                p.close()
                p.join()
                for l,w in m:
                    log_eval+=l*w
                    weight+=w

            log_eval/=weight
            # want to minimize f(x) → send exp(- beta * f(x)) to MH
            # sending log so just - beta f
            self._log_eval=-self.beta*log_eval
            return self._log_eval+offset

    def vary(self, offset=0):
        """makes the parameters vary randomly"""
        self.q_forward, self.q_backward=self.BASE_MOD.vary()
        if hasattr(self, "models"):
            agg=self.BASE_MOD.aggregate()
            for m in self.models:
                m.degregate(agg)
        if hasattr(self, "aggregat"):
            delattr(self, "aggregat")
        if hasattr(self, "_log_eval"):
            delattr(self, "_log_eval")

    def get_q_forward(self):
        return self.q_forward
    def get_q_backward(self):
        return self.q_backward

    def aggregate(self):
        if hasattr(self, "aggregat"):
            if self.aggregat[0] is None or np.isnan(self.aggregat[0]):
                self.aggregat[0]=-self.log_evaluate()/self.beta
            return self.aggregat
        else:
            self.aggregat=np.hstack([-self.log_evaluate()/self.beta,
                                        self.BASE_MOD.aggregate()])
            return self.aggregat

    def degregate(self, aggregat):
        aggregat=aggregat.copy()
        self.aggregat=aggregat
        if aggregat[0] is not None and not np.isnan(aggregat[0]):
            self._log_eval=aggregat[0]
        self.BASE_MOD.degregate(aggregat[1:])
        if hasattr(self, "models"):
            for m in self.models:
                m.degregate(self.aggregat[1:])
        if hasattr(self, "q_forward"):
            delattr(self, "q_forward")
        if hasattr(self, "q_backward"):
            delattr(self, "q_backward")

    def info_num(self, num):
        if num>0:
            return Model.info_num(num-1)
        else:
            return 'estimator', False
