import numpy as np

"""
Different contributions that can be added to the model
"""

# x:
# 0 - G,
# 1 - A_r,
# 2 - G_r,
# 3 - Ca,
# 4 - Cg
# 5 - future A_r/(A_r+G_r)

# Dilution
def dilution(dx, x, t, p):
    dx[0:5]-=p.get_D(t)*x[0:5] # things get diluted
    dx[3:5]+=p.get_D(t)*p.get_Cin(t)

def switching(dx, x, t, p):
    dx[0]+= -p.get_H2()*x[0]  +p.get_H1()*(x[1]+x[2])

    if (x[1]+x[2])!=0:
        prop=x[1]/(x[1]+x[2])
    else:
        prop=p.get_present_p(t,x[5])
    dx[1]+=p.get_H2()*x[0]*prop -p.get_H1()*x[1]
    dx[2]+=p.get_H2()*x[0]*(1-prop) -p.get_H1()*x[2]

def growth(dx, x, t, p):
    # cell ccttion
    mu_g=p.get_mu_mg() * p.get_past_Cg(t,x[4]) / (p.get_Kg()+p.get_past_Cg(t,x[4]))
    mu_a=p.get_mu_ma() * p.get_past_Ca(t,x[3]) / (p.get_Ka()+p.get_past_Ca(t,x[3]))
    dx[0:3]+=mu_g*x[0:3]
    dx[1]+=mu_a*x[1]
    # sugar ccttion -- no delay, consumption "from the future"
    mu_g=p.get_mu_mg() * x[4]/(p.get_Kg()+x[4])
    mu_a=p.get_mu_ma() * x[3]/(p.get_Ka()+x[3])
    dx[4]-= mu_g * np.sum(x[0:3])/p.get_Yxg()
    dx[3]-= mu_a * np.sum(x[1:3])*x[5]/p.get_Yxa()

def single_growth (dx, x, t, p):
    """like growth, but no co-consumption"""
    # cell ccttion
    mu_g=p.get_mu_mg() * p.get_past_Cg(t,x[4]) / (p.get_Kg()+p.get_past_Cg(t,x[4]))
    mu_a=p.get_mu_ma() * p.get_past_Ca(t,x[3]) / (p.get_Ka()+p.get_past_Ca(t,x[3]))
    dx[0]+=mu_g*x[0]
    dx[2]+=mu_g*x[2]
    dx[1]+=mu_a*x[1]
    # sugar ccttion -- no delay, consumption "from the future"
    mu_g=p.get_mu_mg() * x[4]/(p.get_Kg()+x[4])
    mu_a=p.get_mu_ma() * x[3]/(p.get_Ka()+x[3])
    dx[4]-= mu_g * x[0]/p.get_Yxg()
    dx[4]-= mu_g * np.sum(x[1:3])*(1-x[5])/p.get_Yxg()
    dx[3]-= mu_a * np.sum(x[1:3])*x[5]/p.get_Yxa()


# x:
# 0 - G,
# 1 - A_r,
# 2 - G_r,
# 3 - Ca,
# 4 - Cg
# 5 - future A_r/(A_r+G_r)

def follow_p(dx, x, t, p):
    prop=p.get_present_p(t,x[5])
    tot=np.sum(x[1:3])
    da=p.get_present_dp(t,dx[5])*tot

    #if da!=0: # just need to follow the rate from p
              # should never be 0 if prop!=0 and prop!=1
    # else
    if da==0 and prop!=x[1]/np.sum(x[1:3]):
        # p is constant at 0 or 1 and ccttions need to be brought in sync
        if prop==1: # small adjustment needed because growth and
                    # numeric error caused a shift
            da=p.get_ra()*tot
        elif prop==0:   # small adjustment needed because growth and
                        # numeric error caused a shift
            da=-p.get_rd()*tot

    dx[1]+=da
    dx[2]-=da


def react_down(dx, x, t, p):
    if x[5]==1:
        if x[3]>p.get_thresh(): # Ca > thresh
            if p.t_o is not None:
                p.t_o = None
        else: # Ca <= thresh
            if p.t_o is None:
                p.t_o=t
            else:
                # not enough arabinose
                # countdown to reaction was started beforehand
                # if countdown finished, p decreases
                if t>p.t_o+p.get_dtf():
                    dx[5]-=p.get_rd()

def react_up(dx, x, t, p):
    if x[5]==0:
        if x[3]<=p.get_thresh(): # Ca <= thresh
            if p.t_o is not None:
                p.t_o = None
        else: # Ca > thresh
            if p.t_o is None:
                p.t_o=t
            else:
                # not enough arabinose
                # countdown to reaction was started beforehand
                # if countdown finished, p decreases
                if t>p.t_o+p.get_dti():
                    dx[5]+=p.get_ra()
    elif p.get_present_p(t,x[5])==1 and t<p.lock:
        p.lock=t+p.get_dti()/2 # trick to get past the fact solve_ivp may
        # decrease the p even in the future

def react_mid(dx, x, t, p):
    if x[5]>0 and x[5]<1:
        if t<=p.lock:
            dx[5]+=p.get_ra()
        else:
            if x[3]>p.get_thresh(): # Ca > thresh
                dx[5]+=p.get_ra()
            else:
                dx[5]-=p.get_rd()

# x:
# 0 - G,
# 1 - A_r,
# 2 - G_r,
# 3 - Ca,
# 4 - Cg
# 5 - future A_r/(A_r+G_r)

def alt_react(dx, x, t, p):
    """
    use 2 hill functions to determine switching speed between the two
    populations -> 2008Santillan.pdf
    """

    # this expression tends to put the proportions to the hill function as well
    # dA= - r c A + r (1-c) G
    # dA=0 at equilibrium
    # cA = (1-c) G
    # c (A+G)=G

    # only thing missing might be the additional time delays
    # since the original switch is definitely more delayed
    # than the response time for growth

    # rd is used to simplify programming, it doesn't mean rd here
    ex=p.get_thresh()**(p.get_rd())
    # coeff is ~1 when Ca<<thresh
    # in this case, switch from arabinose users to glucose users
    coeff=ex/(ex+p.get_past_Ca(t,x[3])**(p.get_rd()))
    dx[0]+=p.get_ra()*coeff*x[1]
    dx[1]-=p.get_ra()*coeff*x[1]
    dx[0]-=p.get_ra()*(1-coeff)*x[0]
    dx[1]+=p.get_ra()*(1-coeff)*x[0]

def alt_react_2(dx, x, t, p):
    """
    use 2 hill functions to determine switching speed between the two
    populations -> 2008Santillan.pdf
    """

    # this expression tends to put the proportions to the hill function as well
    # dA= - r c A + r (1-c) G
    # dA=0 at equilibrium
    # cA = (1-c) G
    # c (A+G)=G
    # but two thresholds

    # initial test

    # from ara to glu
    t1=.001
    ex=t1**(p.get_rd())
    # coeff is ~1 when Ca<<thresh
    # in this case, switch from arabinose users to glucose users
    coeff=ex/(ex+p.get_past_Ca(t,x[3])**(p.get_rd()))
    dx[0]+=p.get_ra()*coeff*x[1]
    dx[1]-=p.get_ra()*coeff*x[1]

    # from glu to ara
    # t2>t1
    t2=.003
    ex=t2**(p.get_rd())
    coeff=ex/(ex+p.get_past_Ca(t,x[3])**(p.get_rd()))
    dx[0]-=p.get_ra()*(1-coeff)*x[0]
    dx[1]+=p.get_ra()*(1-coeff)*x[0]
