import numpy as np
from numbers import Number
import math
import random

_MAX_EXP=math.log10(2)*np.finfo(float).maxexp
_MIN_EXP=math.log10(2)*np.finfo(float).minexp

class Memory:
    """ Memory class for the models
    Lets the program remember things between iterations and allows to change
    parameters
    """

    LINEAR=[0,1,7,8,9,10,13,14,15]

    NAMES=['dtg', 'dta',
            'Kg', 'Ka',
            'mu_mg', 'mu_ma',
            'thresh',
            'dtf', 'dti',
            'ra', 'rd',
            'H1', 'H2',
            'Yxa', 'Yxg',
            'init_reactive', 'lock_memory']

    def __init__(self, memory=None,
                    # delta t before cell can use glucose
                    dtg=0.1, # h -- supposed
                    # delta t before cell can use arabinose
                    dta=0.5, # h -- time difference between pulse and increase
                    # monod parameters for glucose and arabinose
                    Kg=0.01, # g/L
                    # https://bionumbers.hms.harvard.edu/search.aspx?trm=monod+constant
                    Ka=0.01, # g/L
                    mu_mg=1, # g/L
                    mu_ma=1, # g/L
                    # detection limit by cells for arabinose
                    # should be below detection limit in HPLC → ~0.01 ?
                    # https://www.pnas.org/content/pnas/99/11/7373.full.pdf
                    # → positive from 0.002 g/L and up
                    thresh=0.002, # g/L
                    # time needed before arabinose consumers start going down
                    dtf= None,#2-dta, # h -- in segregostat, ~ time before pop starts
                               # getting under 1000 after pop started decreasing
                    # time needed before arabinose consumers start going up
                    dti=None,#1-dta, # h -- initial response time
                    # rate (h^-1) needed to get from 100% glucose to 100% ara
                    # ascending rate
                    ra=1, # ~100%/h in the initial transition from experiments
                    # rate (h^-1) needed to get from 100% arabinose to 100% glu
                    rd=1,#ra, # unknown?
                    # switching rates between two subpopulations
                    # slow though
                    H1=0.1, # h^-1 # switching from ara to glu
                    H2=0.1, # h^-1
                    # conversion efficiency to biomass
                    # unknown but should be close to 0.5
                    Yxa=0.5, # g biomass/g sugar
                    Yxg=0.5, # g biomass/g sugar
                    # dilution rate -- may be a table if changes across time
                    D=0.5, # h^-1
                    # input concentration -- may be a table
                    # if changes across time
                    Cin= [1.5, 5], # g/L arabinose, glucose, respectively
                    # intermediate variable, can be accessed by the outside
                    t_o=None, lock=True,
                    init_mass=1.8, # g/L
                    # ini=2e6 * 1e6 * 500 * 1e-15 # cells/uL * uL/L * fg/cell * g/fg
                    # https://bionumbers.hms.harvard.edu/bionumber.aspx?id=111211&ver=4&trm=coli+dry+mass&org=
                    # init=2e6 * 1e6 * 280 * 1e-15# cells/uL * uL/L * fg/cell * g/fg
                    # https://bionumbers.hms.harvard.edu/search.aspx?trm=mass+e+coli
                    # ~1.8 g/L at the end of the batch phase (measured)
                    init_reactive=0.9, # part of the initial reactive population
                    t_init=0
                        ):
        self.dtg=dtg
        self.dta=dta
        self.Kg=Kg
        self.Ka=Ka
        self.mu_mg=mu_mg
        self.mu_ma=mu_ma
        self.thresh=thresh
        if dtf is not None:
            self.dtf=dtf
        else:
            self.dtf=2-dta
        if dti is not None:
            self.dti=dti
        else:
            self.dti=1-dta
        self.ra=ra
        self.rd=rd
        self.H1=H1
        self.H2=H2
        self.Yxa=Yxa
        self.Yxg=Yxg
        self.D=D
        self.Cin=np.array(Cin)
        self.t_o=t_o
        self.lock=np.inf
        if not lock:
            self.lock*=-1
        self.lock_memory=lock # just to remember the original value
        self.init_mass=init_mass
        self.init_reactive=init_reactive
        self.t_init=t_init
        # memory:
        # expect a vector of 4 (at time 0) or a 5-column matrix (first is time)
        if memory is not None:
            length_mem=max(100,memory.shape[0])
            self.memory=np.empty((length_mem, 5))
            self.memory.fill(np.nan)
            memory=np.array(memory)
            if len(memory.shape)==1:
                self.memory[0,0]=0
                self.memory[0,1:]=memory
            else:
                self.memory[0:memory.shape[0],:]=memory
        else:
            self.memory=np.empty((1, 5))
            self.memory.fill(np.nan)

    # simple get functions:
    def get_dtg(self):
        return self.dtg
    def get_dta(self):
        return self.dta
    def get_Kg(self):
        return self.Kg
    def get_Ka(self):
        return self.Ka
    def get_mu_mg(self):
        return self.mu_mg
    def get_mu_ma(self):
        return self.mu_ma
    def get_thresh(self):
        return self.thresh
    def get_dtf(self):
        return self.dtf
    def get_dti(self):
        return self.dti
    def get_ra(self):
        return self.ra
    def get_rd(self):
        return self.rd
    def get_H1(self):
        return self.H1
    def get_H2(self):
        return self.H2
    def get_Yxa(self):
        return self.Yxa
    def get_Yxg(self):
        return self.Yxg

    # dtg
    # dta
    # Kg
    # Ka
    # mu_mg
    # mu_ma
    # thresh
    # dtf
    # dti
    # rd >0
    # ra
    # H1
    # H2
    # Yxa
    # Yxg

    # D → time dependent
    # Cin=[Cain, Cgin] → time dependent
    def get_D(self, t=0):
        if isinstance(self.D, Number) or len(self.D.shape)<2:
            # D is constant
            return self.D
        else:
            # D is a [time, dilution rate] 2-column matrix
            # time is the beginning of the dilution rate
            return self.D[np.searchsorted(self.D[:,0], t, side='right')-1, 1]
    def get_Cin(self, t=0):
        if len(self.Cin.shape)==1:
            return self.Cin
        else:
            # 3 column matrix
            return self.Cin[np.searchsorted(self.Cin[:,0], t, side='right')-1, 1:]


    def get_slices(self, t):
        # check if D and/or Cin have no time
        if not hasattr(self,'_slice'):
            D_num=isinstance(self.D, Number) or len(self.D.shape)<2
            C_one=len(self.Cin.shape)==1
            if  D_num and C_one:
                # D and Cin constant
                arr= np.array([t])
            elif D_num and not C_one:
                # only D constant
                arr= self.Cin[:,0]
            elif C_one and not D_num:
                # only Cin constant
                arr= self.D[:,0]
            else:
                # they both have times
                if np.all(self.D[:,0]==self.Cin[:,0]):
                    arr= self.D[:,0]
                # there are differences in the time -> combine arrays
                else:
                    arr=[]
                    d, c=(0,0)
                    while d<self.D.shape[0] and c<self.Cin.shape[0]:
                        if self.D[d,0]==self.Cin[c,0]:
                            arr.append(self.D[d,0])
                        elif self.D[d,0]<self.Cin[c,0]:
                            arr.append(self.D[d,0])
                            d+=1
                        else:
                            arr.append(self.Cin[c,0])
                            c+=1
                    arr=np.array(arr)
                    if   d<self.D.shape[0]:
                        arr=np.hstack(arr, self.D[d:,0])
                    # both can't happen at once due to previous while
                    elif c<self.Cin.shape[0]:
                        arr=np.hstack(arr, self.Cin[c:,0])
            self._slice=arr
        if self._slice[-1]<t:
            arr=np.hstack([self._slice, t])
        else:
            i=np.searchsorted(self._slice, t, side='left')
            arr=self._slice[0:i+1]
        return arr



    def get_init(self):
        # initial conditions:
        #     0    1   2   3    4     5     6
        #   time,  G, A_r, G_r, Ca,   Cg, future A_r/(A_r+G_r)

        # memory:
        # 0    1  2   3   4
        # time,Ca,Cg, p, dp


        init_react=self.init_reactive
        if np.isnan(init_react):
            init_react=0

        # the time to look at should always be the second one from the memory
        # except if only one line
        init=np.zeros(7)
        p=self.get_present_p(t=self.t_init,
            x=self.memory[np.searchsorted(
            self.memory[:,0], self.t_init, side='right')-1,3])
        init[1]=self.init_mass*(1-init_react)
        init[2]=self.init_mass*init_react*p
        init[3]=self.init_mass*init_react*(1-p)
        if len(self.memory.shape)!=1:
            init[0]=self.memory[1,0]
            init[4]=self.memory[1,1]
            init[5]=self.memory[1,2]
            init[6]=self.memory[1,3]
        else:
            init[0]=self.memory[0]
            init[4]=self.memory[1]
            init[5]=self.memory[2]
            init[6]=self.memory[3]

        return init

    # t_o

    # store(dx, x, t) what's needed of dx and x
    def store(self, dx, x, t):
        i=np.searchsorted(self.memory[:,0], t, side='left')
        if i>=self.memory.shape[0]:
            m=np.empty((self.memory.shape[0]*2, self.memory.shape[1]))
            m.fill(np.nan)
            m[0:self.memory.shape[0],:]=self.memory
            self.memory=m
        self.memory[i,0]=t
        x[x<0]=0
        self.memory[i,1:4]=x[3:6]
        self.memory[i,4]=dx[5]
        if x[5]>1:
            self.memory[i,3]=1


    # get_past_Ca(t) - memory 1
    # get_past_Cg(t) - memory 2
    # get_present_p(t) - memory 3
    # get_present_dp(t) - memory 4-- should never be 0 if prop!=0 and prop!=1
    # make sure positive values !

    def _get_past(self, t, delay, index, alt):
        if np.isnan(delay):
            return alt
        t0=t-delay
        i=np.searchsorted(self.memory[:,0], t0, side='right')
        if i<self.memory.shape[0] and not np.isnan(self.memory[i, 0]) :
            k=(t0-self.memory[i-1,0])/(self.memory[i,0]-self.memory[i-1,0])
            return k*self.memory[i, index] + (1-k)*self.memory[i-1,index]
        else:
            if t==self.memory[i-1,0]:
                return alt
            k=(t0-self.memory[i-1,0])/(t-self.memory[i-1,0])
            return (1-k)*self.memory[i-1,index] + k*alt

    def get_past_Ca(self, t, x):
        return self._get_past(t, self.dta, 1, x)

    def get_past_Cg(self, t, x):
        return self._get_past(t, self.dtg, 2, x)

    def get_present_p(self, t, x):
        # dta is the relevant delay for the arabinose proportion
        return self._get_past(t, self.dta, 3, x)

    def get_present_dp(self, t, alt):
        # dta is the relevant delay for the arabinose proportion
        if np.isnan(self.dta):
            return alt

        t0=t-self.dta
        i=np.searchsorted(self.memory[:,0], t, side='right')
        if i<self.memory.shape[0] and not np.isnan(self.memory[i, 0]):

            k=(t-self.memory[i-1,0])/(self.memory[i,0]-self.memory[i-1,0])
            # return all or nothing:
            if k<=0.5:
                return self.memory[i-1,4]
            else:
                return self.memory[i,4]
        else:
            if t==self.memory[i-1,0]:
                return alt
            k=(t0-self.memory[i-1,0])/(t-self.memory[i-1,0])
            # return all or nothing:
            if k<=0.5:
                return self.memory[i-1,4]
            else:
                return alt

    def get_last_time(self):
        return self.memory[np.searchsorted(
            self.memory[:,0], np.nan, side='left')-1, 0]


    ## Functions required for the Metropolis-Hastings algorithm

    CHEMOSTAT=np.array([0.5, 1.5, 5]) # D, Cain, Cgin
    SEGREGOSTAT=np.array([0.5, 0, 5]) # D, Cain, Cgin
    PULSE=np.array([0.5, 7.76e-3]) # mass of arabinose, volume added

    def set_regulation(self, data):
        """set regulation based on a segdata object"""

        if hasattr(self, "_slice"):
            delattr(self, "_slice")

        # base on standard behaviour -- could do something different if needed
        # later, by passing as arguments
        d=data.flow[data.flow[:,2]!="None",2:5]
        # d: [regulation string,
        #     regulation start time,
        #     regulation stop time]
        if d.shape[0]==0:
            # the thing should be just a segregostat
            if not isinstance(self.D, Number) or len(self.D.shape)<2:
                # D is not constant, though it should be
                self.D=Memory.CHEMOSTAT[0]
            if len(self.Cin.shape)!=1:
                # Cin is not constant, although it should be
                self.Cin=Memory.CHEMOSTAT[1:]
        else: # There is regulation noted
        # just replace current regulation because there is no reason not to?
            # D is [time, dilution rate] 2-column matrix
            # Cin is a [time, Cain, Cgin] 3-column matrix
            l=2*d.shape[0]# for the number of times needed
            # if the initial time isn't 0, then the tables should have an
            # additional t=0 line
            if d[0,1]!=0:
                l+=1

            self.D=np.zeros((l,2))
            self.Cin=np.zeros((l,3))

            # no regulation means segregostat
            # chemostat means chemostat
            # until further modification, anything else is a pulse

            # invariant:
            # i is where to add in D and Cin
            # j is where to take time values

            i=0
            if d[0,1]!=0:
                self.D[i, 1]=Memory.SEGREGOSTAT[0]
                self.Cin[i,1:]=Memory.SEGREGOSTAT[1:]
                i+=1

            # for each line of the regulation table
            for j in range(d.shape[0]):
                # if the line is chemostat
                if d[j,0]=="chemostat":
                    self.D[i, 0] =d[j, 1]
                    self.Cin[i,0]=d[j, 1]
                    self.D[i, 1]   =Memory.CHEMOSTAT[0]
                    self.Cin[i, 1:]=Memory.CHEMOSTAT[1:]
                    self.D[i+1, 0] =d[j, 2]
                    self.Cin[i+1,0]=d[j, 2]
                    self.D[i+1, 1]   =Memory.SEGREGOSTAT[0]
                    self.Cin[i+1, 1:]=Memory.SEGREGOSTAT[1:]
                # if the line is a pulse
                else:
                    # need to compute the dilution rate
                    # for exactness' sake
                    # neglected
                    # Dp=Memory.SEGREGOSTAT[0]+Memory.PULSE[1]/(d[j,2]-d[j,1])
                    Dp=Memory.SEGREGOSTAT[0]

                    # SEGREGOSTAT= D, Cain, Cgin
                    # PULSE=mass of arabinose, volume added

                    # Cgin is just diluted a bit
                    Cgin_p=Memory.SEGREGOSTAT[2]*Memory.SEGREGOSTAT[0]/Dp
                    # Cain is technically diluted a bit (from 0)
                    Cain_p=Memory.SEGREGOSTAT[1]*Memory.SEGREGOSTAT[0]/Dp
                    # then there is the mass added in the volume added in the
                    # time difference
                    # technically in g/L since there is one liter of volume in
                    # the reactor

                    Cain_p+= Memory.PULSE[0]/(d[j,2]-d[j,1])/Dp
                        # g /h *h
                    self.D[i, 0] =d[j, 1]
                    self.Cin[i,0]=d[j, 1]
                    self.D[i, 1]   =Dp
                    self.Cin[i, 1:]=[Cain_p, Cgin_p]
                    self.D[i+1, 0] =d[j, 2]
                    self.Cin[i+1,0]=d[j, 2]
                    self.D[i+1, 1]   =Memory.SEGREGOSTAT[0]
                    self.Cin[i+1, 1:]=Memory.SEGREGOSTAT[1:]
                i+=2

    # full set of variable parameters:

    # self.dtg=dtg
    # self.dta=dta
    # self.Kg=Kg
    # self.Ka=Ka
    # self.mu_mg=mu_mg
    # self.mu_ma=mu_ma
    # self.thresh=thresh
    # self.dtf=dtf
    # self.dti=dti
    # self.ra=ra
    # self.rd=rd
    # self.H1=H1
    # self.H2=H2
    # self.Yxa=Yxa
    # self.Yxg=Yxg
    # self.init_reactive=init_reactive

    def aggregate(self):
        """return the set of parameters that change"""
        return np.array([self.dtg,
                        self.dta,
                        self.Kg,
                        self.Ka,
                        self.mu_mg,
                        self.mu_ma,
                        self.thresh,
                        self.dtf,
                        self.dti,
                        self.ra,
                        self.rd,
                        self.H1,
                        self.H2,
                        self.Yxa,
                        self.Yxg,
                        self.init_reactive,
                        self.lock_memory])

    def degregate(self, values, time=0):
        """make the parameters into the set of values referenced"""

        # don't need to erase here, normally, but do it anyways
        # time is for time to erase from
        # also reset lock

        self.dtg=values[0]
        self.dta=values[1]
        self.Kg=values[2]
        self.Ka=values[3]
        self.mu_mg=values[4]
        self.mu_ma=values[5]
        self.thresh=values[6]
        self.dtf=values[7]
        self.dti=values[8]
        self.ra=values[9]
        self.rd=values[10]
        self.H1=values[11]
        self.H2=values[12]
        self.Yxa=values[13]
        self.Yxg=values[14]
        self.init_reactive=values[15]
        self.lock_memory=values[16]
        self.lock=np.inf
        if self.lock_memory:
            self.t_o=None
        else:
            self.lock*=-1
        self.erase(time=time)

    def erase(self, time=0):
        """erase self.memory for all t>time"""
        i=np.searchsorted(self.memory[:,0], time, side='right')
        if i<self.memory.shape[0]:
            if not np.isnan(self.memory[i,0]):
                self.memory[i:,:].fill(np.nan)
        self.lock=np.inf
        if not self.lock_memory:
            lock*=-1

    def clone(self):
        """clone the memory, nothing can be shared"""
        m= Memory(  memory=np.copy(self.memory),
                        # delta t before cell can use glucose
                        dtg=self.dtg, # h -- supposed
                        # delta t before cell can use arabinose
                        dta=self.dta, # h -- time difference between pulse and increase
                        # monod parameters for glucose and arabinose
                        Kg=self.Kg, # g/L
                        # https://bionumbers.hms.harvard.edu/search.aspx?trm=monod+constant
                        Ka=self.Ka, # g/L
                        mu_mg=self.mu_mg, # g/L
                        mu_ma=self.mu_ma, # g/L
                        # detection limit by cells for arabinose
                        # should be below detection limit in HPLC → ~0.01 ?
                        # https://www.pnas.org/content/pnas/99/11/7373.full.pdf
                        # → positive from 0.002 g/L and up
                        thresh=self.thresh, # g/L
                        # time needed before arabinose consumers start going down
                        dtf= self.dtf,#2-dta, # h -- in segregostat, ~ time before pop starts
                                   # getting under 1000 after pop started decreasing
                        # time needed before arabinose consumers start going up
                        dti=self.dti,#1-dta, # h -- initial response time
                        # rate (h^-1) needed to get from 100% glucose to 100% ara
                        # ascending rate
                        ra=self.ra, # ~100%/h in the initial transition from experiments
                        # rate (h^-1) needed to get from 100% arabinose to 100% glu
                        rd=self.rd,#ra, # unknown?
                        # switching rates between two subpopulations
                        # slow though
                        H1=self.H1, # h^-1 # switching from ara to glu
                        H2=self.H2, # h^-1
                        # conversion efficiency to biomass
                        # unknown but should be close to 0.5
                        Yxa=self.Yxa, # g biomass/g sugar
                        Yxg=self.Yxg, # g biomass/g sugar
                        # dilution rate -- may be a table if changes across time
                        D=np.copy(self.D), # h^-1
                        # input concentration -- may be a table
                        # if changes across time
                        Cin= np.copy(self.Cin), # g/L arabinose, glucose, respectively
                        # intermediate variable, can be accessed by the outside
                        t_o=self.t_o, lock=self.lock,
                        init_mass=self.init_mass, # g/L
                        # ini=2e6 * 1e6 * 500 * 1e-15 # cells/uL * uL/L * fg/cell * g/fg
                        # https://bionumbers.hms.harvard.edu/bionumber.aspx?id=111211&ver=4&trm=coli+dry+mass&org=
                        # init=2e6 * 1e6 * 280 * 1e-15# cells/uL * uL/L * fg/cell * g/fg
                        # https://bionumbers.hms.harvard.edu/search.aspx?trm=mass+e+coli
                        # ~1.8 g/L at the end of the batch phase (measured)
                        init_reactive=self.init_reactive, # part of the initial reactive population
                        t_init=self.t_init)
        m.erase(self.t_init)
        if hasattr(self, "_slice"):
            m._slice=self._slice
        return m

    def vary(self, time=0):
        """vary the parameters and erase memory for all t>time
        return q_forward, q_backward
        """

        # erase memory
        self.erase(time)

        # vary parameters
        # don't actually vary the lock, it's there in aggregate for convenience
        forward=np.ones(16)
        backward=np.ones(16)
        self.dtg, forward[0], backward[0]= Memory._vary_linear(self.dtg,d=0.1)
        self.dta, forward[1], backward[1]= Memory._vary_linear(self.dta,d=0.1)
        self.Kg, forward[2], backward[2]=  Memory._vary_expo(self.Kg,min=-5,max=1)
        self.Ka, forward[3], backward[3]=  Memory._vary_expo(self.Ka,min=-5,max=1)
        self.mu_mg, forward[4], backward[4]=Memory._vary_expo(self.mu_mg,max=1)
        self.mu_ma, forward[5], backward[5]=Memory._vary_expo(self.mu_ma,max=1)
        self.thresh, forward[6], backward[6]=Memory._vary_expo(self.thresh, max=0)
        self.dtf, forward[7], backward[7]=Memory._vary_linear(self.dtf, d=0.1)
        self.dti, forward[8], backward[8]=Memory._vary_linear(self.dti, d=0.1)
        self.ra, forward[9], backward[9]= Memory._vary_linear(self.ra)
        self.rd, forward[10], backward[10]=Memory._vary_linear(self.rd)
        self.H1, forward[11], backward[11]=Memory._vary_expo(self.H1)
        self.H2, forward[12], backward[12]=Memory._vary_expo(self.H2)
        self.Yxa, forward[13], backward[13]=Memory._vary_linear(self.Yxa, max=1)
        self.Yxg, forward[14], backward[14]=Memory._vary_linear(self.Yxg, max=1)
        self.init_reactive, forward[15], backward[15]= Memory._vary_linear(
                                                self.init_reactive, max=1)
        forward=np.log(forward)
        backward=np.log(backward)

        return np.sum(forward), np.sum(backward)

    def info_num(num):
        """return whether parameter was varied linearly, and its name"""
        return Memory.NAMES[num], num in Memory.LINEAR

    def _vary_expo(val, min=_MIN_EXP, max=_MAX_EXP, d=1):
        """Vary value according to an exponential rule
        Basically the log_10 of the values varies linearly
        returns partial q_forward and q_backward as well
        """
        if np.isnan(val):
            return val, 1, 1
        v, f, b = Memory._vary_linear(math.log10(val), min=min, max=max,d=d)
        return math.pow(10,v), f, b


    def _vary_linear(val, min=0, max=np.finfo(float).max, d=1):
        """Vary value according to linear rule
        d should be 10* the order of magnitude of the variation
        """
        if np.isnan(val):
            return val, 1, 1
        delta=max-min
        if delta>d:
            delta=d
        inc=d/10 # the max variation on either side of the value
        interval=[val-inc, val+inc]
        if interval[0]<min:
            interval[0]=min
        # if min is too low, max can't be too high, given the construction of
        # delta
        elif interval[1]>max:
            interval[1]=max

        # make random number in interval
        new=interval[0]+random.random()*(interval[1]-interval[0])

        # uniform distribution over the interval
        forward=1/(interval[1]-interval[0])

        # compute backward proba:
        # construction of the interval based on new:
        interval=[new-inc, new+inc]
        if interval[0]<min:
            interval[0]=min
        # if min is too low, max can't be too high, given the construction of
        # delta
        elif interval[1]>max:
            interval[1]=max

        backward=1/(interval[1]-interval[0])

        return new, forward, backward
