import numpy as np
from scipy.integrate import solve_ivp
import tensorflow as tf
from tensorflow import keras
from . import Memory

class Model:
    """
    The model class
    Create with Memory
    then run
    """

    # solution structure, in lines:
    # 0: time
    # 1: Cells using only G
    # 2: Responsive cells using A
    # 3: Responsive cells using only G
    # 4: Ara concentration
    # 5: Glucose concentration
    # 6: future A_r/(A_r+G_r)

    def __init__(self, memory, diff):
        self.memory=memory
        self.diff=diff
        init=memory.get_init()
        # make it a vertical array for solutions
        if len(init.shape)==1:
            init=init.reshape(init.shape[0], 1)
        self.sol=init # first line is times

    def run(self, t1):
        times=self.memory.get_slices(t1)
        for t in times:
            # run the model in slices → allows bigger steps
            self._run(t)

    def _run(self, t1):
        """Run the model from the last time in memory to t1"""
        if t1>self.sol[0,-1]:
            new_sol=solve_ivp(self.diff, [self.sol[0,-1], t1],
                                self.sol[1:,-1],
                                args=(self.memory,), max_step=0.1) # 1e-4 to not miss a pulse

            # solve_ivp gets a horizontal table
            if self.sol.shape[1]>1:
                self.sol=np.hstack(( self.sol[:,:-1],
                                     np.vstack((new_sol.t, new_sol.y))  ))
            else:
                self.sol=np.vstack((new_sol.t, new_sol.y))

    def clone(self):
        """clone the model, deep copy without results
        erase required memory from memory as well
        """
        return Model(self.memory.clone(), self.diff)

    def set_regulation(self, data):
        """from a segdata object, set the regulation in the memory"""
        self.memory.set_regulation(data)

    def compare(self, data):
        """compare the current model with the segdata object provided
        return logcosh loss and number of elements compared, in that order
        """
        # no need to remember things here as in theory it should be remembered
        # in the MetroWrapper

        # if time required bigger than current max time
        if data.flow[-1,0]>self.sol[0,-1]:
            self.run(data.flow[-1,0])
            # print("done with run")

        # compute the values of the models where needed
        # could be integrated into run but for now easier to put here

        # could use searchsorted for each iteration but wasteful

        # only compare values>0
        d=np.searchsorted(data.flow[:,0], 0.0) # for data
        if data.flow[d,0]<0.0:
            d+=1
        s=0 # for self

        # mod_values has the values of the model
        # d_values has the values from data
        mod_values=np.empty_like(data.flow[:,0])
        d_values=np.empty_like(data.flow[:,0])
        while d<data.flow.shape[0]:

            if s<self.sol.shape[1]:
                if data.flow[d,0]>self.sol[0,s]:
                    s+=np.searchsorted(self.sol[0,s:], data.flow[d,0])


            if s<self.sol.shape[1]:
                # factor to compute the value at t[d]
                prop=( (self.sol[0,s] -data.flow[d-1,0])/
                       (data.flow[d,0]-data.flow[d-1,0]) )
                # model value for proportion of GFP+ at t[s]
                m_s= self.sol[2,s]/np.sum(self.sol[1:4,s])
            else:
                prop=0
                m_s=0
            # model value for proportion of GFP+ at t[s-1]
            m_s1= self.sol[2,s-1]/np.sum(self.sol[1:4,s-1])

            mod_values[d]=(prop*m_s+
                        (1-prop)*m_s1)

            d_values[d]=data.flow[d,5].count_proportion()

            # invariant: t[s-1]<t[d]<=t[s]
            d+=1

        # print("done extracting")

        # compute the logcosh
        # using keras for easy computation and control of overflow
        # may relatively easily add other comparison methods later
        return keras.losses.logcosh(d_values.astype('float32'),
                                    mod_values.astype('float32')
                                    ).numpy(), data.flow.shape[0]


    def aggregate(self):
        """return the parameters that change, in a list"""
        return self.memory.aggregate()

    def degregate(self, values):
        """makes the model into the one defined by the parameter values"""
        # remove the results of the model
        self.memory.degregate(values, time=self._erase())

    def vary(self):
        """change the model slightly, return q_forward and q_backward"""
        return self.memory.vary(time=self._erase())

    def _erase(self):
        self.sol=self.sol[:,0]
        self.sol=self.sol.reshape(self.sol.shape[0], 1)
        self.memory.erase(time=self.sol[0,0])
        return self.sol[0,0]

    def info_num(num):
        return Memory.info_num(num)

    def plot_prop(self, ax, label=''):
        prop=self.sol[2,:]/np.sum(self.sol[1:4,:], axis=0)
        ax.plot(self.sol[0,:], 100*prop, label=label)
